# SV comparison pipeline
import logging
import os
import shutil
import subprocess
import sys
from os.path import join as pjoin

import pandas as pd
from simppl.simple_pipeline import SimplePipeline
from ugbio_core.exec_utils import print_and_execute
from ugbio_core.logger import logger
from ugbio_core.vcfbed import vcftools

from ugbio_comparison.vcf_pipeline_utils import VcfPipelineUtils


class SVComparison:
    """
    SV comparison pipeline
    """

    def __init__(self, simple_pipeline: SimplePipeline | None = None, logger: logging.Logger | None = None):
        """Combines VCF in parts from GATK and indices the result

        Parameters
        ----------
        simple_pipeline : SimplePipeline, optional
            Optional SimplePipeline object for executing shell commands
        """
        self.sp = simple_pipeline
        self.vpu = VcfPipelineUtils(self.sp)
        if logger is None:
            self.logger = logging.getLogger(__name__)
        else:
            self.logger = logger

    def __execute(self, command: str, output_file: str | None = None):
        """Summary

        Parameters
        ----------
        command : str
            Description
        output_file : str, optional
            Description
        """
        print_and_execute(command, output_file=output_file, simple_pipeline=self.sp, module_name=__name__)

    def collapse_vcf(
        self,
        vcf: str,
        output_vcf: str,
        bed: str | None = None,
        pctseq: float = 0.0,
        pctsize: float = 0.0,
    ):
        """
        Collapse VCF using truvari collapse

        Parameters
        ----------
        vcf : str
            Input VCF file
        output_vcf : str
            Output VCF file
        bed : str, optional
            Bed file, by default None
        pctseq : float, optional
            Percentage of sequence identity, by default 0.0
        pctsize : float, optional
            Percentage of size identity, by default 0.0

        Returns
        -------
        None
        """

        truvari_cmd = ["truvari", "collapse", "-i", vcf, "--passonly", "-t"]

        if bed:
            truvari_cmd.extend(["--bed", bed])
        truvari_cmd.extend(["--pctseq", str(pctseq)])
        truvari_cmd.extend(["--pctsize", str(pctsize)])

        self.logger.info(f"truvari command: {' '.join(truvari_cmd)}")
        p1 = subprocess.Popen(truvari_cmd, stdout=subprocess.PIPE)
        p2 = subprocess.Popen(["bcftools", "view", "-Oz", "-o", output_vcf], stdin=p1.stdout)  # noqa: S607
        p1.stdout.close()  # Allow p1 to receive a SIGPIPE if p2 exits.
        p2.communicate()  # Wait for p2 to finish
        p1.wait()  # Wait for p1 to finish
        if p1.returncode != 0:
            raise RuntimeError(f"truvari collapse failed with error code {p1.returncode}")
        if p2.returncode != 0:
            raise RuntimeError(f"bcftools view failed with error code {p2.returncode}")

        removed_vcf_path = "removed.vcf"  # Parameterize the file path
        if os.path.exists(removed_vcf_path):
            os.unlink(removed_vcf_path)
            self.logger.info(f"Deleted temporary file: {removed_vcf_path}")
        else:
            self.logger.warning(f"Temporary file not found: {removed_vcf_path}")

    def run_truvari(
        self,
        calls: str,
        gt: str,
        outdir: str,
        bed: str | None = None,
        pctseq: float = 0.0,
        pctsize: float = 0.0,
        *,
        erase_outdir: bool = True,
    ):
        """
        Run truvari, generate truvari report and concordance VCF
        Parameters
        ----------
        calls : str
            Calls file
        gt : str
            Ground truth file
        outdir : str
            Output directory
        bed : str, optional
            Bed file, by default None
        pctseq : float, optional
            Percentage of sequence identity, by default 0.0
        pctsize : float, optional
            Percentage of size identity, by default 0.0
        erase_outdir : bool, optional
            Erase output directory if it exists, by default True

        Returns
        -------
        None
        """
        if erase_outdir and os.path.exists(outdir):
            shutil.rmtree(outdir)

        truvari_cmd = [
            "truvari",
            "bench",
            "-b",
            gt,
            "-c",
            calls,
            "-o",
            outdir,
            "-t",
            "--passonly",
        ]

        if bed:
            truvari_cmd.extend(["--includebed", bed])
        truvari_cmd.extend(["--pctseq", str(pctseq)])
        truvari_cmd.extend(["--pctsize", str(pctsize)])

        self.logger.info(f"truvari command: {' '.join(truvari_cmd)}")
        self.__execute(" ".join(truvari_cmd))

    def truvari_to_dataframes(
        self,
        truvari_dir: str,
        custom_info_fields: tuple = (),
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Convert truvari report to dataframes

        Parameters
        ----------
        truvari_dir : str
            Truvari directory
        custom_info_fields : tuple, optional
            Custom info fields to read from the VCFs, in addition SVTYPE and SVLEN

        Returns
        -------
        tuple[pd.DataFrame, pd.DataFrame]
            Tuple of truvari base and calls concordance dataframes
        """
        info_fields_to_read: list[str] = list(("SVTYPE", "SVLEN") + custom_info_fields)
        df_tp_base = vcftools.get_vcf_df(pjoin(truvari_dir, "tp-base.vcf.gz"), custom_info_fields=info_fields_to_read)
        df_tp_base["svlen"] = df_tp_base["svlen"].apply(lambda x: x[0] if isinstance(x, tuple) else x).fillna(0)
        df_tp_base["label"] = "TP"
        df_fn = vcftools.get_vcf_df(pjoin(truvari_dir, "fn.vcf.gz"), custom_info_fields=info_fields_to_read)
        df_fn["svlen"] = df_fn["svlen"].apply(lambda x: x[0] if isinstance(x, tuple) else x).fillna(0)
        df_fn["label"] = "FN"
        df_base = pd.concat((df_tp_base, df_fn))

        df_tp_calls = vcftools.get_vcf_df(pjoin(truvari_dir, "tp-comp.vcf.gz"), custom_info_fields=info_fields_to_read)
        df_tp_calls["label"] = "TP"
        df_tp_calls["svlen"] = df_tp_calls["svlen"].apply(lambda x: x[0] if isinstance(x, tuple) else x).fillna(0)

        df_fp = vcftools.get_vcf_df(pjoin(truvari_dir, "fp.vcf.gz"), custom_info_fields=info_fields_to_read)
        df_fp["label"] = "FP"
        df_fp["svlen"] = df_fp["svlen"].apply(lambda x: x[0] if isinstance(x, tuple) else x).fillna(0)

        df_calls = pd.concat((df_tp_calls, df_fp))
        return df_base, df_calls

    def run_pipeline(
        self,
        calls: str,
        gt: str,
        output_file_name: str,
        outdir: str,
        hcr_bed: str | None = None,
        pctseq: float = 0.0,
        pctsize: float = 0.0,
        custom_info_fields: tuple = (),
        *,
        erase_outdir: bool = True,
    ):
        """
        Run truvari pipeline

        Parameters
        ----------
        calls : str
            Calls file
        gt : str
            Ground truth file
        output_file_name : str
            Name of the output H5 concordance file
        outdir : str
            Output directory
        hcr_bed : str, optional
            High confidence region bed file, by default None
        pctseq : float, optional
            Percentage of sequence identity, by default 0.0
        pctsize : float, optional
            Percentage of size identity, by default 0.0
        erase_outdir : bool, optional
            Erase output directory if it exists, by default True
        custom_info_fields : list[str], optional
            Custom info fields to read from the VCFs, in addition SVTYPE and SVLEN

        Returns
        -------
        None
        """
        self.logger.info(f"Running truvari pipeline with calls: {calls} and gt: {gt}")
        calls_fn = calls
        tmpfiles_to_move = []
        self.collapse_vcf(
            calls_fn,
            calls_fn.replace(".vcf.gz", "_collapsed.vcf.gz"),
            bed=hcr_bed,
            pctseq=pctseq,
            pctsize=pctsize,
        )
        calls_fn = calls_fn.replace(".vcf.gz", "_collapsed.vcf.gz")
        tmpfiles_to_move.append(calls_fn)

        self.vpu.sort_vcf(calls_fn, calls_fn.replace("_collapsed.vcf.gz", "_collapsed.sort.vcf.gz"))
        calls_fn = calls_fn.replace("_collapsed.vcf.gz", "_collapsed.sort.vcf.gz")
        tmpfiles_to_move.append(calls_fn)
        tmpfiles_to_move.append(calls_fn + ".tbi")

        self.vpu.index_vcf(calls_fn)

        gt_fn = gt

        self.collapse_vcf(
            gt_fn,
            gt_fn.replace(".vcf.gz", "_collapsed.vcf.gz"),
            bed=hcr_bed,
            pctseq=pctseq,
            pctsize=pctsize,
        )
        gt_fn = gt_fn.replace(".vcf.gz", "_collapsed.vcf.gz")
        tmpfiles_to_move.append(gt_fn)
        self.vpu.sort_vcf(gt_fn, gt_fn.replace("_collapsed.vcf.gz", "_collapsed.sort.vcf.gz"))
        gt_fn = gt_fn.replace("_collapsed.vcf.gz", "_collapsed.sort.vcf.gz")
        tmpfiles_to_move.append(gt_fn)
        tmpfiles_to_move.append(gt_fn + ".tbi")
        self.vpu.index_vcf(gt_fn)

        self.run_truvari(
            calls=calls_fn,
            gt=gt_fn,
            outdir=outdir,
            bed=hcr_bed,
            pctseq=pctseq,
            pctsize=pctsize,
            erase_outdir=erase_outdir,
        )
        df_base, df_calls = self.truvari_to_dataframes(outdir, custom_info_fields=custom_info_fields)
        df_base.to_hdf(output_file_name, key="base", mode="w")
        df_calls.to_hdf(output_file_name, key="calls", mode="a")
        for tmpfile in tmpfiles_to_move:
            if os.path.exists(tmpfile):
                shutil.move(tmpfile, outdir)
        self.logger.info(f"truvari pipeline finished with calls: {calls_fn} and gt: {gt_fn}")


def get_parser():
    """
    Get argument parser for SVComparison

    Returns
    -------
    argparse.ArgumentParser
        Argument parser
    """
    import argparse

    parser = argparse.ArgumentParser(description="SV Comparison Pipeline")
    parser.add_argument("--calls", required=True, help="Input calls VCF file")
    parser.add_argument("--gt", required=True, help="Input ground truth VCF file")
    parser.add_argument("--output_filename", required=True, help="output h5 with concordance file")
    parser.add_argument("--outdir", required=True, help="Full path to output dir to TRUVARI results")
    parser.add_argument("--hcr_bed", help="High confidence region bed file")
    parser.add_argument("--pctseq", type=float, default=0.0, help="Percentage of sequence identity")
    parser.add_argument("--pctsize", type=float, default=0.0, help="Percentage of size identity")
    parser.add_argument("--custom_info_fields", nargs="+", default=[], help="Custom info fields to read from the VCFs")
    parser.add_argument("--verbosity", default="INFO", help="Logging verbosity level")
    return parser


def run(argv):
    parser = get_parser()
    SimplePipeline.add_parse_args(parser)
    args = parser.parse_args(argv[1:])
    logger.setLevel(getattr(logging, args.verbosity))
    sp = SimplePipeline(args.fc, args.lc, debug=args.d, print_timing=True)
    pipeline = SVComparison(sp, logger)
    pipeline.run_pipeline(
        calls=args.calls,
        gt=args.gt,
        hcr_bed=args.hcr_bed,
        pctseq=args.pctseq,
        pctsize=args.pctsize,
        outdir=args.outdir,
        output_file_name=args.output_filename,
        custom_info_fields=tuple(args.custom_info_fields),
    )


def main():
    run(sys.argv)


if __name__ == "__main__":
    main()



from collections import defaultdict

import numpy as np
import pandas as pd
import pyfaidx
import pysam
import ugbio_core.concordance.flow_based_concordance as fbc
import ugbio_core.vcfbed.variant_annotation as annotation
from ugbio_core.consts import DEFAULT_FLOW_ORDER
from ugbio_core.logger import logger
from ugbio_core.vcfbed import bed_writer, vcftools


def _fix_errors(df: pd.DataFrame) -> pd.DataFrame:
    """Parses dataframe generated from the VCFEVAL concordance VCF and prepares it for
     classify/classify_gt functions that only consider the genotypes of the call and the base
    only rather than at the classification that VCFEVAL produced

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe

    Returns
    -------
    pd.DataFrame
        Output dataframe,
    """

    # remove genotypes of variants that were filtered out and thus are false negatives
    # (VCFEVAL outputs UG genotype for ignored genotypes too and they are classified downstream
    # as true positives if we do not make this modification)
    fix_tp_fn_loc = (df["call"] == "IGN") & ((df["base"] == "FN") | (df["base"] == "FN_CA"))
    replace = df.loc[fix_tp_fn_loc, "gt_ultima"].apply(lambda x: (None,))
    df.loc[replace.index, "gt_ultima"] = replace

    # fix all the places in which vcfeval returns a good result, but the genotype is not adequate
    # in these cases we change the genotype of the gt to be adequate with the classify function as follow:
    # (TP,TP), (TP,None) - should put the values of ultima in the gt
    df.loc[(df["call"] == "TP") & ((df["base"] == "TP") | (df["base"].isna())), "gt_ground_truth"] = df[
        (df["call"] == "TP") & ((df["base"] == "TP") | (df["base"].isna()))
    ]["gt_ultima"]

    # (FP_CA,FN_CA), (FP_CA,None) - Fake a genotype from ultima such that one of the alleles is the same (and only one)
    df.loc[(df["call"] == "FP_CA") & ((df["base"] == "FN_CA") | (df["base"].isna())), "gt_ground_truth"] = df[
        (df["call"] == "FP_CA") & ((df["base"] == "FN_CA") | (df["base"].isna()))
    ]["gt_ultima"].apply(
        lambda x: ((x[0], x[0]) if (len(x) < 2 or (x[1] == 0)) else ((x[1], x[1]) if (x[0] == 0) else (x[0], 0)))  # noqa PLR2004
    )
    return df


def __map_variant_to_dict(variant: pysam.VariantRecord) -> defaultdict:
    """Converts a line from vcfeval concordance VCF to a dictionary. The following fields are extracted
    call genotype, base genotype, qual, chromosome, position, ref, alt and all values from the INFO column

    Parameters
    ----------
    variant : pysam.VariantRecord
        VCF record

    Returns
    -------
    defaultdict
        Output dictionary
    """
    call_sample_ind = 1
    gtr_sample_ind = 0

    return defaultdict(
        lambda: None,
        variant.info.items()
        + [
            ("GT_ULTIMA", variant.samples[call_sample_ind]["GT"]),
            ("GT_GROUND_TRUTH", variant.samples[gtr_sample_ind]["GT"]),
            ("QUAL", variant.qual),
            ("CHROM", variant.chrom),
            ("POS", variant.pos),
            ("REF", variant.ref),
            ("ALLELES", variant.alleles),
        ],
    )


def vcf2concordance(  # noqa PLR0915 C901
    raw_calls_file: str,
    concordance_file: str,
    chromosome: str | None = None,
    scoring_field: str | None = None,
) -> pd.DataFrame:
    """Generates concordance dataframe

    Parameters
    ----------
    raw_calls_file : str
        File with GATK calls (.vcf.gz)
    concordance_file : str
        GenotypeConcordance/VCFEVAL output file (.vcf.gz)
    chromosome: str
        Fetch a specific chromosome (Default - all)
    scoring_field : str, optional
        The name of the INFO field that is used to score the variants.
        This value replaces the TREE_SCORE in the output data frame.
        When None TREE_SCORE is not replaced (default: None)

    No Longer Returned
    ------------------
    pd.DataFrame
    """

    if chromosome is None:
        concord_vcf = pysam.VariantFile(concordance_file)
    else:
        concord_vcf = pysam.VariantFile(concordance_file).fetch(chromosome)

    def call_filter(x):
        # Remove variants that were ignored (either outside of comparison intervals or
        # filtered out).
        return not (
            (x["CALL"] in {"IGN", "OUT"} and x["BASE"] is None)
            or (x["CALL"] in {"IGN", "OUT"} and x["BASE"] in {"IGN", "OUT"})
            or (x["CALL"] is None and x["BASE"] in {"IGN", "OUT"})
        )

    concord_vcf_extend = filter(call_filter, (__map_variant_to_dict(variant) for variant in concord_vcf))

    columns = [
        "CHROM",
        "POS",
        "QUAL",
        "REF",
        "ALLELES",
        "GT_ULTIMA",
        "GT_GROUND_TRUTH",
        "SYNC",
        "CALL",
        "BASE",
        "STR",
        "RU",
        "RPA",
    ]
    column_names = [x.lower() for x in columns]
    concordance_df = pd.DataFrame([[x[y] for y in columns] for x in concord_vcf_extend], columns=column_names)

    # make the gt_ground_truth compatible with GC
    concordance_df["gt_ground_truth"] = concordance_df["gt_ground_truth"].map(
        lambda x: (None, None) if x == (None,) else x
    )

    concordance_df["indel"] = concordance_df["alleles"].apply(lambda x: len({len(y) for y in x}) > 1)
    concordance_df = _fix_errors(concordance_df)

    def classify(x: defaultdict | pd.Series | dict) -> str:
        """Classify a record as true positive / false positive / false negative by matching the alleles.
        TP will be called if some of the alleles match

        Parameters
        ----------
        x : defaultdict, pd.Series or dict
            Input record

        Returns
        -------
        str
            classification
        """
        if x["gt_ultima"] == (None, None) or x["gt_ultima"] == (None,):
            return "fn"

        if x["gt_ground_truth"] == (None, None) or x["gt_ground_truth"] == (None,):
            return "fp"

        # If both gt_ultima and gt_ground_truth are not none:
        set_gtr = set(x["gt_ground_truth"]) - {0}
        set_ultima = set(x["gt_ultima"]) - {0}

        if len(set_gtr & set_ultima) > 0:
            return "tp"

        if len(set_ultima - set_gtr) > 0:
            return "fp"

        # If it is not tp or fp, then return fn:
        return "fn"

    concordance_df["classify"] = concordance_df.apply(classify, axis=1, result_type="reduce")

    def classify_gt(x: defaultdict | pd.Series | dict) -> str:
        """Classify a record as true positive / false negative / false positive. True positive requires
        match in the alleles and genotypes

        Parameters
        ----------
        x : defaultdict, pd.Series or dict
            Input record

        Returns
        -------
        str
            Classification
        """
        n_ref_gtr = len([y for y in x["gt_ground_truth"] if y == 0])
        n_ref_ultima = len([y for y in x["gt_ultima"] if y == 0])

        if x["gt_ultima"] == (None, None) or x["gt_ultima"] == (None,):
            return "fn"
        if x["gt_ground_truth"] == (None, None) or x["gt_ground_truth"] == (None,):
            return "fp"
        if n_ref_gtr < n_ref_ultima:
            return "fn"
        if n_ref_gtr > n_ref_ultima:
            return "fp"
        if x["gt_ultima"] != x["gt_ground_truth"]:
            return "fp"
        # If not fn or fp due to the reasons above:
        return "tp"

    concordance_df["classify_gt"] = concordance_df.apply(classify_gt, axis=1, result_type="reduce")
    concordance_df.loc[
        (concordance_df["classify_gt"] == "tp") & (concordance_df["classify"] == "fp"),
        "classify_gt",
    ] = "fp"

    # cases where we called wrong allele and then filtered out - are false negatives, not false positives
    called_fn = (concordance_df["base"] == "FN") | (concordance_df["base"] == "FN_CA")
    marked_fp = concordance_df["classify"] == "fp"
    concordance_df.loc[called_fn & marked_fp, "classify"] = "fn"
    marked_fp = concordance_df["classify_gt"] == "fp"
    concordance_df.loc[called_fn & marked_fp, "classify_gt"] = "fn"
    concordance_df.index = pd.Index(list(zip(concordance_df.chrom, concordance_df.pos, strict=False)))
    original = vcftools.get_vcf_df(raw_calls_file, chromosome=chromosome, scoring_field=scoring_field)

    concordance_df = concordance_df.drop("qual", axis=1)

    drop_candidates = ["chrom", "pos", "alleles", "indel", "ref", "str", "ru", "rpa"]
    if original.shape[0] > 0:
        concordance = concordance_df.join(
            original.drop(
                [x for x in drop_candidates if x in original.columns and x in concordance_df.columns],
                axis=1,
            )
        )
    else:
        concordance = concordance_df.copy()

        tmp = original.drop(
            [x for x in drop_candidates if x in original.columns and x in concordance_df.columns],
            axis=1,
        )
        for t in tmp.columns:
            concordance[t] = None

    only_ref = concordance["alleles"].apply(len) == 1
    concordance = concordance[~only_ref]

    # Marking as false negative the variants that appear in concordance but not in the
    # original VCF (even if they do show some genotype)
    missing_variants = concordance.index.difference(original.index)
    logger.info("Identified %i variants missing in the input VCF", len(missing_variants))
    missing_variants_non_fn = concordance.loc[missing_variants].query("classify!='fn'").index
    logger.warning(
        "Identified %i variants missing in the input VCF and not marked false negatives",
        len(missing_variants_non_fn),
    )
    concordance.loc[missing_variants_non_fn, "classify"] = "fn"
    concordance.loc[missing_variants_non_fn, "classify_gt"] = "fn"

    return concordance


def bed_file_length(input_bed: str) -> int:
    """Calc the number of bases in a bed file

    Parameters
    ----------
    input_bed : str
        Input Bed file

    Return
    ------
    int
        number of bases in a bed file
    """

    input_df = pd.read_csv(input_bed, sep="\t", header=None)
    input_df = input_df.iloc[:, [0, 1, 2]]
    input_df.columns = ["chr", "pos_start", "pos_end"]
    return np.sum(input_df["pos_end"] - input_df["pos_start"] + 1)


def close_to_hmer_run(
    df: pd.DataFrame,
    runfile: str,
    min_hmer_run_length: int = 10,
    max_distance: int = 10,
) -> pd.DataFrame:
    """Adds column is_close_to_hmer_run and inside_hmer_run that is T/F"""
    df["close_to_hmer_run"] = False
    df["inside_hmer_run"] = False
    run_df = bed_writer.parse_intervals_file(runfile, min_hmer_run_length)
    gdf = df.groupby("chrom")
    grun_df = run_df.groupby("chromosome")
    for chrom in gdf.groups.keys():
        gdf_ix = gdf.groups[chrom]
        grun_ix = grun_df.groups[chrom]
        pos1 = np.array(df.loc[gdf_ix, "pos"])
        pos2 = np.array(run_df.loc[grun_ix, "start"])
        pos1_closest_pos2_start = np.searchsorted(pos2, pos1) - 1
        close_dist = abs(pos1 - pos2[np.clip(pos1_closest_pos2_start, 0, None)]) < max_distance
        close_dist |= abs(pos2[np.clip(pos1_closest_pos2_start + 1, None, len(pos2) - 1)] - pos1) < max_distance
        pos2 = np.array(run_df.loc[grun_ix, "end"])
        pos1_closest_pos2_end = np.searchsorted(pos2, pos1)
        close_dist |= abs(pos1 - pos2[np.clip(pos1_closest_pos2_end - 1, 0, None)]) < max_distance
        close_dist |= abs(pos2[np.clip(pos1_closest_pos2_end, None, len(pos2) - 1)] - pos1) < max_distance
        is_inside = pos1_closest_pos2_start == pos1_closest_pos2_end
        df.loc[gdf_ix, "inside_hmer_run"] = is_inside
        df.loc[gdf_ix, "close_to_hmer_run"] = close_dist & (~is_inside)
    return df


def annotate_concordance(
    concordance_df: pd.DataFrame,
    fasta: str,
    bw_high_quality: list[str] | None = None,
    bw_all_quality: list[str] | None = None,
    annotate_intervals: list[str] | None = None,
    runfile: str | None = None,
    flow_order: str | None = DEFAULT_FLOW_ORDER,
    hmer_run_length_dist: tuple = (10, 10),
) -> tuple[pd.DataFrame, list]:
    """Annotates concordance data with information about SNP/INDELs and motifs

    Parameters
    ----------
    concordance_df : pd.DataFrame
        Concordance dataframe
    fasta : str
        Indexed FASTA of the reference genome
    bw_high_quality : list[str], optional
        Coverage bigWig file from high mapq reads  (Optional)
    bw_all_quality : list[str], optional
        Coverage bigWig file from all mapq reads  (Optional)
    annotate_intervals : list[str], optional
        Interval files for annotation
    runfile : str, optional
        bed file with positions of hmer runs (in order to mark homopolymer runs)
    flow_order : str, optional
        Flow order
    hmer_run_length_dist : tuple, optional
        tuple (min_hmer_run_length, max_distance) for marking variants near homopolymer runs

    No Longer Returned
    ------------------
    pd.DataFrame
        Annotated dataframe
    list
        list of the names of the annotations
    """

    if annotate_intervals is None:
        annotate_intervals = []

    logger.info("Marking SNP/INDEL")
    concordance_df = annotation.classify_indel(concordance_df)
    logger.info("Marking H-INDEL")
    concordance_df = annotation.is_hmer_indel(concordance_df, fasta)
    logger.info("Marking motifs")
    concordance_df = annotation.get_motif_around(concordance_df, 5, fasta)
    logger.info("Marking GC content")
    concordance_df = annotation.get_gc_content(concordance_df, 10, fasta)
    if bw_all_quality is not None and bw_high_quality is not None:
        logger.info("Calculating coverage")
        concordance_df = annotation.get_coverage(concordance_df, bw_high_quality, bw_all_quality)
    if runfile is not None:
        length, dist = hmer_run_length_dist
        logger.info("Marking homopolymer runs")
        concordance_df = close_to_hmer_run(concordance_df, runfile, min_hmer_run_length=length, max_distance=dist)
    annots = []
    if annotate_intervals is not None:
        for annotation_file in annotate_intervals:
            logger.info("Annotating intervals")
            concordance_df, annot = annotation.annotate_intervals(concordance_df, annotation_file)
            annots.append(annot)
    logger.debug("Filling filter column")  # debug since not interesting step
    concordance_df = annotation.fill_filter_column(concordance_df)

    logger.info("Filling filter column")
    if flow_order is not None:
        concordance_df = annotation.annotate_cycle_skip(concordance_df, flow_order=flow_order)
    return concordance_df, annots


def reinterpret_variants(
    concordance_df: pd.DataFrame,
    reference_fasta: str,
    *,
    ignore_low_quality_fps: bool = False,
) -> pd.DataFrame:
    """Reinterprets the variants by comparing the variant to the ground truth in flow space

    Parameters
    ----------
    concordance_df : pd.DataFrame
        Input dataframe
    reference_fasta : str
        Indexed FASTA
    ignore_low_quality_fps : bool, optional
        Shoud the low quality false positives be ignored in reinterpretation (True for mutect, default False)

    See Also
    --------
    `flow_based_concordance.py`

    No Longer Returned
    ------------------
    pd.DataFrame
        Reinterpreted dataframe
    """
    logger.info("Variants reinterpret")
    concordance_df_result = pd.DataFrame()
    fasta = pyfaidx.Fasta(reference_fasta, build_index=False, rebuild=False)
    for contig in concordance_df["chrom"].unique():
        concordance_df_contig = concordance_df.loc[concordance_df["chrom"] == contig]
        input_dict = _get_locations_to_work_on(concordance_df_contig, ignore_low_quality_fps=ignore_low_quality_fps)
        concordance_df_contig = fbc.reinterpret_variants(concordance_df_contig, input_dict, fasta)
        concordance_df_result = pd.concat([concordance_df_result, concordance_df_contig])
    return concordance_df_result


def _get_locations_to_work_on(input_df: pd.DataFrame, *, ignore_low_quality_fps: bool = False) -> dict:
    """Dictionary of  in the dataframe that we care about

    Parameters
    ----------
    input_df : pd.DataFrame
        Input
    ignore_low_quality_fps : bool, optional
        Should we ignore the low quality false positives

    Returns
    -------
    dict
        locations dictionary split between fps/fns/tps etc.

    """
    filtered_df = vcftools.FilterWrapper(input_df)
    fps = filtered_df.reset().get_fp().get_df()
    if "tree_score" in fps.columns and fps["tree_score"].dtype == np.float64 and ignore_low_quality_fps:
        cutoff = fps.tree_score.quantile(0.80)
        fps = fps.query(f"tree_score > {cutoff}")
    fns = filtered_df.reset().get_df().query('classify=="fn"')
    tps = filtered_df.reset().get_tp().get_df()
    gtr = (
        filtered_df.reset()
        .get_df()
        .loc[filtered_df.get_df()["gt_ground_truth"].apply(lambda x: x not in [(None, None), (None,)])]
        .copy()
    )
    gtr = gtr.sort_values("pos")
    ugi = (
        filtered_df.reset()
        .get_df()
        .loc[filtered_df.get_df()["gt_ultima"].apply(lambda x: x not in [(None, None), (None,)])]
        .copy()
    )
    ugi = ugi.sort_values("pos")

    pos_fps = np.array(fps.pos)
    pos_gtr = np.array(gtr.pos)
    pos_ugi = np.array(ugi.pos)
    pos_fns = np.array(fns.pos)

    result = {
        "fps": fps,
        "fns": fns,
        "tps": tps,
        "gtr": gtr,
        "ugi": ugi,
        "pos_fps": pos_fps,
        "pos_gtr": pos_gtr,
        "pos_ugi": pos_ugi,
        "pos_fns": pos_fns,
    }

    return result



#!/env/python
# Copyright 2022 Ultima Genomics Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
# DESCRIPTION
#    Compare UG callset to ground truth (preferentially using VCFEVAL as an engine)
# CHANGELOG in reverse chronological order

from __future__ import annotations

import argparse
import logging
import os
import sys
from shutil import copyfile

import pandas as pd
import pysam
from joblib import Parallel, delayed
from simppl.simple_pipeline import SimplePipeline
from tqdm import tqdm
from ugbio_core.consts import DEFAULT_FLOW_ORDER
from ugbio_core.h5_utils import read_hdf
from ugbio_core.logger import logger
from ugbio_core.vcfbed import vcftools
from ugbio_core.vcfbed.interval_file import IntervalFile

from ugbio_comparison import comparison_utils, vcf_pipeline_utils
from ugbio_comparison.comparison_pipeline import ComparisonPipeline

MIN_CONTIG_LENGTH = 100000


def _contig_concordance_annotate_reinterpretation(  # noqa: PLR0913
    # pylint: disable=too-many-arguments
    raw_calls_vcf,
    concordance_vcf,
    contig,
    reference,
    bw_high_quality,
    bw_all_quality,
    annotate_intervals,
    runs_intervals,
    hpol_filter_length_dist,
    flow_order,
    base_name_outputfile,
    disable_reinterpretation,
    ignore_low_quality_fps,
    scoring_field,
):
    logger.info("Reading %s", contig)
    concordance = comparison_utils.vcf2concordance(
        raw_calls_vcf,
        concordance_vcf,
        contig,
        scoring_field=scoring_field,
    )

    annotated_concordance, _ = comparison_utils.annotate_concordance(
        concordance,
        reference,
        bw_high_quality,
        bw_all_quality,
        annotate_intervals,
        runs_intervals,
        hmer_run_length_dist=hpol_filter_length_dist,
        flow_order=flow_order,
    )

    if not disable_reinterpretation:
        annotated_concordance = comparison_utils.reinterpret_variants(
            annotated_concordance, reference, ignore_low_quality_fps=ignore_low_quality_fps
        )
    logger.debug("%s: %s", contig, annotated_concordance.shape)
    annotated_concordance.to_hdf(f"{base_name_outputfile}{contig}.h5", key=contig)


def get_parser() -> argparse.ArgumentParser:
    ap_var = argparse.ArgumentParser(prog="run_comparison_pipeline.py", description="Compare VCF to ground truth")
    ap_var.add_argument(
        "--n_parts",
        help="Number of parts that the VCF is split into",
        required=True,
        type=int,
    )
    ap_var.add_argument("--input_prefix", help="Prefix of the input file", required=True, type=str)
    ap_var.add_argument("--output_file", help="Output h5 file", required=True, type=str)
    ap_var.add_argument(
        "--output_interval",
        help="Output bed file of intersected intervals",
        required=True,
        type=str,
    )
    ap_var.add_argument("--gtr_vcf", help="Ground truth VCF file", required=True, type=str)
    ap_var.add_argument(
        "--cmp_intervals",
        help="Ranges on which to perform comparison (bed/interval_list)",
        required=False,
        type=str,
        default=None,
    )
    ap_var.add_argument(
        "--highconf_intervals",
        help="High confidence intervals (bed/interval_list)",
        required=True,
        type=str,
    )
    ap_var.add_argument(
        "--runs_intervals",
        help="Runs intervals (bed/interval_list)",
        required=False,
        type=str,
        default=None,
    )
    ap_var.add_argument(
        "--annotate_intervals",
        help="interval files for annotation (multiple possible)",
        required=False,
        type=str,
        default=None,
        action="append",
    )
    ap_var.add_argument("--reference", help="Reference genome", required=True, type=str)
    ap_var.add_argument("--reference_dict", help="Reference genome dictionary", required=False, type=str)
    ap_var.add_argument(
        "--coverage_bw_high_quality",
        help="BigWig file with coverage only on high mapq reads",
        required=False,
        default=None,
        type=str,
        action="append",
    )
    ap_var.add_argument(
        "--coverage_bw_all_quality",
        help="BigWig file with coverage on all mapq reads",
        required=False,
        default=None,
        type=str,
        action="append",
    )
    ap_var.add_argument(
        "--call_sample_name",
        help="Name of the call sample",
        required=True,
        default="sm1",
    )
    ap_var.add_argument("--truth_sample_name", help="Name of the truth sample", required=True)
    ap_var.add_argument("--header_file", help="Desired header", required=False, default=None)
    ap_var.add_argument(
        "--filter_runs",
        help="Should variants on hmer runs be filtered out",
        action="store_true",
    )
    ap_var.add_argument(
        "--hpol_filter_length_dist",
        nargs=2,
        type=int,
        help="Length and distance to the hpol run to mark",
        default=[10, 10],
    )
    ap_var.add_argument("--ignore_filter_status", help="Ignore variant filter status", action="store_true")
    ap_var.add_argument(
        "--revert_hom_ref",
        help="For DeepVariant callsets - revert filtered hom_ref to het_ref for max recall calculation",
        action="store_true",
    )
    ap_var.add_argument(
        "--scoring_field",
        help="The pipeline expects a TREE_SCORE column in order to score the variants. If another field is \
        provided via scoring_field then its values will be copied to the TREE_SCORE column",
        required=False,
        default=None,
        type=str,
    )
    ap_var.add_argument(
        "--flow_order",
        type=str,
        help="Sequencing flow order (4 cycle)",
        required=False,
        default=DEFAULT_FLOW_ORDER,
    )
    ap_var.add_argument(
        "--output_suffix",
        help="Add suffix to the output file",
        required=False,
        default="",
        type=str,
    )
    ap_var.add_argument(
        "--disable_reinterpretation",
        help="Should re-interpretation be run",
        action="store_true",
    )
    ap_var.add_argument(
        "--special_chromosome",
        help="The chromosome that would be used for the \
        'concordance' dataframe (whole genome mode only)",
        default="chr9",
    )
    ap_var.add_argument("--is_mutect", help="Are the VCFs output of Mutect (false)", action="store_true")
    ap_var.add_argument("--n_jobs", help="n_jobs of parallel on contigs", type=int, default=-1)
    ap_var.add_argument(
        "--verbosity",
        help="Verbosity: ERROR, WARNING, INFO, DEBUG",
        required=False,
        default="INFO",
    )
    return ap_var


def run(argv: list[str]):
    """Concordance between VCF and ground truth"""
    parser = get_parser()
    SimplePipeline.add_parse_args(parser)
    args = parser.parse_args(argv[1:])
    logger.setLevel(getattr(logging, args.verbosity))
    sp = SimplePipeline(args.fc, args.lc, debug=args.d, print_timing=True)
    vpu = vcf_pipeline_utils.VcfPipelineUtils(sp)

    cmp_intervals = IntervalFile(sp, args.cmp_intervals, args.reference, args.reference_dict)
    highconf_intervals = IntervalFile(sp, args.highconf_intervals, args.reference, args.reference_dict)
    runs_intervals = IntervalFile(sp, args.runs_intervals, args.reference, args.reference_dict)

    # intersect intervals and output as a bed file
    if cmp_intervals.is_none():  # interval of highconf_intervals
        logger.info(f"copy {args.highconf_intervals} to {args.output_interval}")
        copyfile(highconf_intervals.as_bed_file(), args.output_interval)
    else:
        vpu.intersect_bed_files(cmp_intervals.as_bed_file(), highconf_intervals.as_bed_file(), args.output_interval)

    args_dict = {k: str(vars(args)[k]) for k in vars(args)}
    pd.DataFrame(args_dict, index=[0]).to_hdf(args.output_file, key="input_args")

    comparison_pipeline = ComparisonPipeline(
        vpu=vpu,
        n_parts=args.n_parts,
        input_prefix=args.input_prefix,
        truth_file=args.gtr_vcf,
        cmp_intervals=cmp_intervals,
        highconf_intervals=highconf_intervals,
        ref_genome=args.reference,
        call_sample=args.call_sample_name,
        truth_sample=args.truth_sample_name,
        output_file_name=args.output_file,
        header=args.header_file,
        output_suffix=args.output_suffix,
        ignore_filter=args.ignore_filter_status,
        revert_hom_ref=args.revert_hom_ref,
    )
    raw_calls_vcf, concordance_vcf = comparison_pipeline.run()

    # single interval-file concordance - will be saved in a single dataframe

    if not cmp_intervals.is_none():
        concordance_df = comparison_utils.vcf2concordance(
            raw_calls_vcf,
            concordance_vcf,
            scoring_field=args.scoring_field,
        )
        annotated_concordance_df, _ = comparison_utils.annotate_concordance(
            concordance_df,
            args.reference,
            args.coverage_bw_high_quality,
            args.coverage_bw_all_quality,
            args.annotate_intervals,
            runs_intervals.as_bed_file(),
            hmer_run_length_dist=args.hpol_filter_length_dist,
            flow_order=args.flow_order,
        )

        if not args.disable_reinterpretation:
            annotated_concordance_df = comparison_utils.reinterpret_variants(
                annotated_concordance_df,
                args.reference,
                ignore_low_quality_fps=args.is_mutect,
            )

        annotated_concordance_df.to_hdf(args.output_file, key="concordance", mode="a")
        # hack until we totally remove chr9
        annotated_concordance_df.to_hdf(args.output_file, key="comparison_result", mode="a")
        vcftools.bed_files_output(
            annotated_concordance_df,
            args.output_file,
            mode="w",
            create_gt_diff=(not args.is_mutect),
        )

    # whole-genome concordance - will be saved in dataframe per chromosome
    else:
        with pysam.VariantFile(raw_calls_vcf) as variant_file:
            # we filter out short contigs to prevent huge files
            contigs = [
                x for x in variant_file.header.contigs if variant_file.header.contigs[x].length > MIN_CONTIG_LENGTH
            ]

        base_name_outputfile = os.path.splitext(args.output_file)[0]
        Parallel(n_jobs=args.n_jobs, max_nbytes=None)(
            delayed(_contig_concordance_annotate_reinterpretation)(
                raw_calls_vcf,
                concordance_vcf,
                contig,
                args.reference,
                args.coverage_bw_high_quality,
                args.coverage_bw_all_quality,
                args.annotate_intervals,
                runs_intervals.as_bed_file(),
                args.hpol_filter_length_dist,
                args.flow_order,
                base_name_outputfile,
                args.disable_reinterpretation,
                args.is_mutect,
                args.scoring_field,
            )
            for contig in tqdm(contigs)
        )

        # merge temp h5 files

        # find columns and set the same header for empty dataframes
        df_columns = None
        for contig in contigs:
            h5_temp = read_hdf(f"{base_name_outputfile}{contig}.h5", key=contig)
            if h5_temp.shape == (0, 0):  # empty dataframes are dropped to save space
                continue
            df_columns = pd.DataFrame(columns=h5_temp.columns)
            break

        for contig in contigs:
            h5_temp = read_hdf(f"{base_name_outputfile}{contig}.h5", key=contig)
            if h5_temp.shape == (0, 0):  # empty dataframes get default columns
                h5_temp = pd.concat((h5_temp, df_columns), axis=1)
            h5_temp.to_hdf(args.output_file, mode="a", key=contig)
            if contig == args.special_chromosome:
                h5_temp.to_hdf(args.output_file, mode="a", key="concordance")
            os.remove(f"{base_name_outputfile}{contig}.h5")

        write_mode = "w"
        for contig in contigs:
            annotated_concordance_df = read_hdf(args.output_file, key=contig)
            vcftools.bed_files_output(
                annotated_concordance_df,
                args.output_file,
                mode=write_mode,
                create_gt_diff=(not args.is_mutect),
            )
            write_mode = "a"


def main():
    run(sys.argv)


if __name__ == "__main__":
    main()
